/*

Exif.js: Exif properties editor
================================================================================

Änderung von JPEG Metadaten. Die beschreibenden Daten werden im Exif-Format
gespeichert (siehe Exchangeable Image File Format)

Updating JPEG metadata stored as Exif tags. For Exif see also Exchangeable Image
File Format in public sources.

================================================================================


* Tags of interest:

DateTime 	   0x0132
Title 			0x010E
MSTITLE			0x9C9B
Comment   		0x9C9C
Manual tags		0x9C9E
Subject   		0x9C9F
Copyright		0x8298

ExifSpecId     0x8769
GPSSpecId      0x8825
InteropIFD     0xA005


Copyright Hartmut V. Bornemann, 2018
*
* */

#feature-id    Utilities > EXIF

#include <pjsr/DataType.jsh>
#include <pjsr/StdButton.jsh>
#include <pjsr/FrameStyle.jsh>
#include <pjsr/TextAlign.jsh>
#include <pjsr/ButtonCodes.jsh>
#include <pjsr/NumericControl.jsh>
#include <pjsr/StdDialogCode.jsh>
#include <pjsr/FileMode.jsh>

#define SOI  0xD8
#define SOS  0xDA
#define APP0 0xE0
#define APP1 0xE1
#define APP2 0xE2
#define DQT  0xDB
#define SET_PREFIX "Exif"

#define TITLE "EXIF - Exif Properties Editor"
#define VERSION "1.0.0"

#define test 0

#define ExifSpecId      0x8769
#define GPSSpecId       0x8825
#define InteropIFD      0xA005

#define TagID_DateTime 	0x0132
#define MYTITLE 			0x010E
#define MSTITLE			0x9C9B
#define MYCOMMENTS 		0x9C9C
#define MANUALTAGS 		0x9C9E
#define MYSUBJECT 		0x9C9F
#define MYCOPYRIGHT		0x8298

// global
var isLittleEndian = false;


function isSpecIFD(tagId)
{
	return [ExifSpecId, GPSSpecId, InteropIFD].indexOf(tagId) > -1;
}

function showDialog()
{
   this.__base__ = Dialog;
   this.__base__();
 	this.userResizable = false;
   var dialog 			= this;
	var exifInfo		= null;
	var exifBlock 		= null;
	var IFDs 			= null;
	var image 			= null;
	var pictureBox 	= null;
	var tbxFilename 	= null;
	var btnUpdate		= null;
 	var btnRecent		= null;

	var editableTags = [
		new exifTag(MYTITLE  		, 2, "Title"),				//  10E
		new exifTag(MYSUBJECT		, 1, "Subject"),			// 9C9F
		new exifTag(MYCOMMENTS		, 1, "Comments"),			// 9C9C
		new exifTag(MYCOPYRIGHT		, 2, "Copyright"),		// 8298
		new exifTag(315  				, 2, "Authors"),			// 013B
		new exifTag(TagID_DateTime	, 2, "Date taken"),
		new exifTag(305  				, 2, "Program name"),
		new exifTag(271  				, 2, "Camera maker"),
		new exifTag(272  				, 2, "Camera model"),
		new exifTag(MANUALTAGS		, 1, "Tags")
	];

   this.lblCopyright = new Label(this)
   this.lblCopyright.text = "© 2017, Hartmut V. Bornemann";

	this.editFields = [];

	for (var i = 0; i < editableTags.length; i++)
	{
		var tag = editableTags[i];
		var lbl = new Label(this);
		lbl.text = editableTags[i].key;
		var txt = new Edit(this);
		txt.enabled = false;
		txt.setMinWidth(250);

		if (tag.tagId == TagID_DateTime)
			txt.toolTip ="<p>The format is YYYY-MM-DD HH:MM:SS with time shown " +
				"in 24-hour format, and the date and time separated by one blank.</p>";
		else if (tag.tagId == MANUALTAGS)
			txt.toolTip ="<p>Comma separated tags (i.e.: Galaxy;Star Trails.</p>";

		tag.edit = txt;
		var hoz  = new HorizontalSizer(this);
		hoz.spacing = 20;
		hoz.add(lbl);
		hoz.addStretch();
		hoz.add(txt);
		this.editFields.push(hoz);
	}

	this.gbx1 = new GroupBox( this );
	with (this.gbx1)
	{
		title = "Image properties";
		sizer = new VerticalSizer;
		sizer.margin = 8;
		sizer.spacing = 8;

 		for (var i = 0; i < this.editFields.length; i++)
		{
			sizer.add( this.editFields[i] );
		}
	}

	this.tbxFilename = new Edit(this);
	this.tbxFilename.readOnly = true;
	tbxFilename = this.tbxFilename;
	this.btnOpenFile = new PushButton(this);
	with (this.btnOpenFile)
	{
		text = "Open";
		toolTip = "Open a JPEG file";

#ifneq __PI_PLATFORM__ MACOSX
    	icon = this.scaledResource ( ":/browser/select-file.png" );
#endif

		onPress = function()
		{
			var dlg = new OpenFileDialog();
			with (dlg)
			{
				caption = "open a jpeg image file";
				multipleSelections = false;
				filters = [["JPEG", ".jpg"]];
				if (execute() == StdDialogCode_Ok)
				{
					try{
						image = new Bitmap( fileName );
						pictureBox.repaint();
					}
					catch(err) {
    					Console.writeln(err);
					}
  					tbxFilename.text = fileName;

					exifInfo = readAPP1(fileName, editableTags);

					if (exifInfo == null) return;

					exifBlock = exifInfo.exifBlock;
					IFDs = exifInfo.IFDs;

					btnUpdate.enabled = true;
					btnRecent.enabled = true;
				}
			}
		}
	}

	this.btnRecent   = new PushButton(this);
	with (this.btnRecent)
	{
		icon = this.scaledResource( ":/icons/download.png" );
		enabled = false;
		text = "recent";
		toolTip = "Load properties from last updated file";
		onPress = function()
		{
			//
			// get defaults for empty keys
			//
			var allTags = serializeIFDs(IFDs);
			for (var i = 0; i < allTags.length; i++)
			{
				var tag = allTags[i];
				if (tag.tagId == TagID_DateTime) continue;
				if (tag.edit != null)
				{
					var v = getSetting(tag.key);
					if (v != "") tag.setString( v );
				}
			}
		}
	}

	btnRecent = this.btnRecent;

	this.btnUpdate   = new PushButton(this);
	with (this.btnUpdate)
	{
		enabled = false;
		icon = this.scaledResource( ":/icons/ok.png" );
		text = "Update";
		toolTip = "Update file with properties";
		onPress = function()
		{
 			//
			// insert new exif
			//
			// overwrite 9C9B (MS tagId for Comments)
			var allTags = serializeIFDs(IFDs);
			var msTitle = getTagByIndex(allTags, MSTITLE);
			if (msTitle != null)
			{
				var myTitle = getTagByIndex(allTags, MYTITLE);
				var text = myTitle.getString();
 				msTitle.setString(text);
			}

			var file = new File(exifInfo.fileName, FileMode_Read|FileMode_Write);
			file.position = exifBlock.remainPos;
			var remains = file.read( DataType_ByteArray, exifBlock.remainSize);
			file.position = exifBlock.offset;

			if (exifBlock.dataLen == 0)
 				Console.writeln("Insert Exif tags @ " + toLHex(file.position, 4));
			else
				Console.writeln("Update Exif tags @ " + toLHex(file.position, 4));

			var buffer = getAPP1Buffer(file, IFDs);

			Console.writeln("APP1 length " + toLHex(buffer.length, 4));

			if (!test)
			{
				file.write(buffer);
				file.write(remains);
			}
			else
				dumpByteArray(buffer, 0);

			file.close();
			Console.writeln("Exif settings up to date");
 			//
			// save settings
			//
			if (!test)
			{
				for (var i = 0; i < allTags.length; i++)
				{
					var tag = allTags[i];
					if (tag.edit != null)
					{
						if (tag.tagId != TagID_DateTime)
						{
							saveSetting(tag.key, tag.getString());
						}
					}
				}
			}

			var allTags = serializeIFDs(IFDs);
			printTags(allTags);
			//
			// clean unwanted editable tags
			//
			var newArray = [];
			for (var i in editableTags)
			{
				if (editableTags[i].edit != null) newArray.push(editableTags[i]);
			}
			editableTags = newArray;
			//
			// re-open
			//
			exifInfo = readAPP1(exifInfo.fileName, editableTags);

			if (exifInfo == null) return;

			exifBlock = exifInfo.exifBlock;
			IFDs = exifInfo.IFDs;
		}
	}

	btnUpdate = this.btnUpdate;

	this.btnExit     = new PushButton(this);
	with (this.btnExit)
	{
		icon = this.scaledResource( ":/icons/cancel.png" );
		text = "Exit";
		toolTip = "Exit";
		onPress = function()
		{
			dialog.ok();
		}
	}

	this.buttonSizer = new HorizontalSizer(this);
	with (this.buttonSizer)
	{
		margin = 8;
		spacing = 8;
		add(this.btnOpenFile);
		addStretch();
		add(this.btnRecent);
		addStretch();
		add(this.btnUpdate);
		addStretch();
		add(this.btnExit);
	}

	this.gbx2 = new GroupBox( this );
	with (this.gbx2)
	{
		title = "File";
		sizer = new VerticalSizer;
		sizer.margin = 8;
		sizer.spacing = 8;
	   sizer.add( this.tbxFilename );
		sizer.add( this.buttonSizer );
	}

	this.pbx = new Frame(this);
	with (this.pbx)
	{
		backgroundColor = 0xFF222222;

		onPaint = function(x0, y0, x1, y1 )
		{
			if (image == null) return;
			var fImage = image.width / image.height;
			var fPict  = width / height;
			var scale = 1;
			if (fImage >= fPict)
				scale = width / image.width;
			else
				scale = height / image.height;
			var img = image.scaled( scale );
			var g = new Graphics( this );
			var p = new Point((width - img.width)/2, (height - img.height)/2);
			var r = new Rect (0, 0, img.width,img.height);
			g.drawBitmapRect( p, img, r );
			g.end();
		}
	}

	pictureBox = this.pbx;

	this.plnLeft = new Frame(this)
	with (this.plnLeft)
	{
		sizer = new HorizontalSizer();
		with (sizer)
		{
	   	margin = 0;
	   	spacing = 0;
			add(this.pbx);
		}
      onResize = function(wNew, hNew, wOld, hOld )
      {
			this.parent.pbx.setScaledFixedWidth( hNew / 3 * 4 );
      }
	}

	this.pnlRight = new Frame(this)
	with (this.pnlRight)
	{
		sizer = new VerticalSizer();
   	with ( sizer ) {
      margin = 8;
      spacing = 8;
		add( this.lblCopyright );
      add( this.gbx1 );
      add( this.gbx2 );
   }
	}

	// all together
   this.sizer = new HorizontalSizer(this)
	with ( this.sizer ) {
	   margin = 8;
	   spacing = 8;
	  	add( this.plnLeft );
	  	add( this.pnlRight );
	}

   // window
   this.windowTitle = TITLE + " " + VERSION;
   this.adjustToContents();
 	this.setFixedSize();
}

// subs

function printTags(tags)
{
   if (!test) return;
	Console.writeln("ALL TAGS ------------------------------------ "+tags.length);
	for (var i = 0; i < tags.length; i++)
	{
		var tag = tags[i];
		if (isSpecIFD (tag.tagId ))
		{
			Console.writeln(i+'\t' + toLHex(tag.tagPtr,4)+'\t'+toLHex(tag.tagId, 4)+'\t' +
								toLHex(tag.dataEndPtr,4));
			//printTags(tag.data);
		}
		else
		{
			if (tag.type == 1)
			{
				Console.writeln(i+'\t' +toLHex(tag.tagPtr,4)+'\t' +toLHex(tag.bufferPtr,4)+'\t'+ "tagId "+toLHex(tag.tagId, 4)+
				'\t'+ tag.type +'\t'+"value:"+'\t' +
								toLHex(tag.dataEndPtr,4)+'\t'+ "data = "+tag.data.length+'\t'+
								tag.getString());
			}
			else if (tag.type == 2)
			{
				Console.write(i+'\t' +toLHex(tag.tagPtr,4)+'\t' +toLHex(tag.bufferPtr,4)+'\t'+ "tagId "+toLHex(tag.tagId, 4)+
				'\t'+ tag.type +'\t'+toLHex(tag.data.length,4)+'\t' );
				Console.writeln(tag.getString () );
			}
			else
				{
					if (tag.n < 5)
					{
						Console.write(i+'\t' +toLHex(tag.tagPtr,4)+'\t' +toLHex(tag.bufferPtr,4)+'\t'+ "tagId "+toLHex(tag.tagId, 4)+
								'\t'+ tag.type +'\t'+toLHex(tag.n, 4) + " bytes" +'\t' +
								toLHex(tag.data.length,4)+" ");
						if (tag.data != null)
							for (var j = 0; j < 4; j++) Console.write(toHex(tag.data.at(j))+" ");
						Console.writeln();
					}
					else
					{
						Console.writeln(i+'\t' +toLHex(tag.tagPtr,4)+'\t' +toLHex(tag.bufferPtr,4)+'\t'+ "tagId "+toLHex(tag.tagId, 4)+
								'\t'+ tag.type +'\t'+toLHex(tag.n, 4) + " bytes" +'\t' +
								toLHex(tag.data.length,4));
					}
				}
		}
	}
}

function getTagByIndex(tags, tagId)
{
  	for (var i in tags) {if (tags[i].tagId == tagId) return tags[i];}
	return null;
}

function readAPP1(fileName, editableTags)
{
	// reset editor text
	for (var i in editableTags) editableTags[i].setString("");
	// open file
 	var file = new File(fileName, FileMode_Read);
	Console.writeln("File " + fileName);
 	// find APP1
	var coll = new collectTags(file);
	var exifBlock = coll.exifBlock;

	if (exifBlock == null)
	{
		file.close();
		message("Bad file: no Exif structure to changed or edit.");
		return null;
	}

	var IFDs = coll.IFDs;

	if (IFDs == null)
	{
		// build IFDs from editableTags
		IFDs = [editableTags];
		for (var i in editableTags)
		{
			var edit = editableTags[i];
			if (edit.tagId == TagID_DateTime)
				edit.setString(dateToExifString(fileName));
		}
	}

	var allTags = serializeIFDs(IFDs);
	printTags(allTags);

	for (var i in editableTags)
	{
 		var edit = editableTags[i];
		var fileTag = getTagByIndex(allTags, edit.tagId);
		if (fileTag == null)
		{
			//
			//append editableTags[i] to IFD0 ExifSpecId
			//
			//Console.writeln("append "+toLHex(edit.tagId,4));
			appendNewTag(IFDs[0], edit);
			//Console.writeln("attach "+toLHex(edit.tagId,4));
			edit.attach(edit);	// attach to itself enables the editor
		}
		else
		{
			//Console.writeln("attach "+toLHex(edit.tagId,4));
			fileTag.attach(edit);
		}
	}

	checkDate(IFDs, fileName);

	var allTags = serializeIFDs(IFDs);
	printTags(allTags);

	file.close();
	return {fileName : fileName, exifBlock : exifBlock, IFDs : IFDs};
}

function appendNewTag(IFD, newTag)
{
	if (newTag.tagId == TagID_DateTime)
	{
		IFD.push(newTag);
		return;
	}
	else
	{
		// Add tag to spec tags
		for (var j in IFD)
		{
			var tag = IFD[j];
			if (tag.tagId == ExifSpecId)
			{
				Console.writeln("newTag "+toLHex(newTag.tagId,4)+ " to " +toLHex(tag.tagId,4));
				tag.data.push(newTag);
				return;
			}
		}
	}
	// add to IFD
	IFD.push(newTag);
	Console.writeln("newTag "+toLHex(newTag.tagId,4)+ " added as # " + IFD.length);
}

function checkDate(IFDs, fileName)
{
	//
	// check, if date is empty
	//
	var allTags = serializeIFDs(IFDs);
	var dateTag = getTagByIndex(allTags, TagID_DateTime);
	var tag0x9003 = getTagByIndex(allTags, 0x9003);
 	if (dateTag.data.length == 0)
	{
 		if (tag0x9003 != null)
		{
			dateTag.setString(tag0x9003.getString());
		}
		else
		{
			dateTag.setString(dateToExifString(fileName));
		}
	}
	//
	// check, if tags 0x9003 & 0x9004 exist
	// if not, create an Exif specific tag and add tags here
	//

	// display & edit date in ISO format, save with Exif-Format
	var dateISO = dateStringToISO(dateTag.getString());
	dateISO = dateISO.replace('T', ' ');
 	dateTag.setString(dateISO);
	// append tag0x9003, if not found (MS Explorer shows this tag)
	if (tag0x9003 == null)
	{
		var specTag = getTagByIndex(allTags, ExifSpecId);
		if (specTag == null)
		{
			specTag = new exifTag(ExifSpecId, 4, "SpecTag");
			specTag.count = 1;
			specTag.n = 4;
			specTag.data = [];
			specTag.isLittleEndian = dateTag.isLittleEndian;
			IFDs[0].push(specTag);
		}
		var tag = new exifTag(0x9003, 2, "TAG0x9003");
		tag.setString(dateTag.getString());
		tag.isLittleEndian = dateTag.isLittleEndian;
		specTag.data.push(tag);
	}
}

function getAPP1Buffer(file, IFDs)
{
	var position = file.position;
	var buffer  = fillBuffer(file, IFDs);
	var APP1Len = 8 + buffer.length;
	var hdr     = new ByteArray();
	addBytesToArray(hdr, [0xFF, 0xE1]);
	hdr.add( numberToByteArray(APP1Len, 2, false));
	if (isLittleEndian)
		addBytesToArray(hdr, [0x45, 0x78, 0x69, 0x66, 0x00, 0x00]);
	else
		addBytesToArray(hdr, [0x45, 0x78, 0x69, 0x66, 0x00, 0x00]);
	hdr.add(buffer);
	return hdr;
}

function fillBuffer(file, IFDs)
{
	//
	var buffer = new ByteArray();
	//
	// build APP1
	//
	// APP1 length (=0) , ExifIdentifier Code and pointer to 1st IFD
	if (isLittleEndian)
		addBytesToArray(buffer, [0x49, 0x49, 0x2A, 0x00]);
	else
		addBytesToArray(buffer, [0x4D, 0x4D, 0x00, 0x2A]);
	//
	// append IFDs
	//
	var nextIFD = buffer.length + 4;
	var prevIFD = 0;		//
	buffer.add( numberToByteArray(nextIFD, 4, isLittleEndian) );
	for (var i in IFDs)
	{
		// ifd is a collection of tags
		var ifd = IFDs[i];

		if (prevIFD > 0)
		{
			var n = numberToByteArray(buffer.length, 4, isLittleEndian);
			for (var j = 0; j < 4; j++) buffer.at(prevIFD + j, n.at(j));
		}
		//
		// position of next pointer, tags counter + n tags x 12
		//
		prevIFD = buffer.length + 2 + ifd.length * 12;
		appendIFD(buffer, ifd);

 		if (i == 1)
		{
			// check image
			var imagePos = 0;
			var tag = ifd[ifd.length - 1];
			file.position = tag.dataEndPtr;
			// find FF D8
			for (;;)
			{
				if (imagePos == 0)
				{
					var x = file.read(DataType_ByteArray, 1);
					if (x.at(0x00) == 0xFF)
					{
						var y = file.read(DataType_ByteArray, 1);
						if (y.at(0x00) == 0xD8)
						{
							imagePos = file.position - 2;
 						}
						else if (y.at(0x00) == 0xD9) break;
					}
				}
				else
				{
					var x = file.read(DataType_ByteArray, 1);
					if (x.at(0x00) == 0xFF)
					{
						var y = file.read(DataType_ByteArray, 1);
						if (y.at(0x00) == 0xD9)
						{
							var l = file.position - imagePos;
							file.position = imagePos;
							var image = file.read(DataType_ByteArray, l);
 							buffer.add(image);
 							break;
						}
					}
				}
			}
		}
	}
	// finished
	return buffer;
}

function appendIFD(buffer, tags)
{
	//
	// set count of tags in collection
	//
	buffer.add (numberToByteArray(tags.length, 2, isLittleEndian));
	//
	// add the tags
	//
	for (var i in tags)
	{
		var tag = tags[i];
		tag.addTag(buffer);
	}
	//
	// add a 4-byte zero offset to the next IFD here
	//
	buffer.add(new ByteArray(4));
	//
	// check Exif specific IDs
	//
	for (var i in tags)
	{
		var tag = tags[i];
		if (isSpecIFD (tag.tagId ))
		{
			copyTo (buffer, tag.bufferPtr + 8, topOf(buffer, isLittleEndian));
			appendIFD(buffer, tag.data);
		}
	}
	//
	// add data at current position, i.e., following the pointer to next IFD
	//
	appendData(buffer, tags);
}

function appendData(buffer, tags)
{
	//
	// copy long (> 4) data to next buffer location
	//
	for (var i = 0; i < tags.length; i++)
	{
		var tag = tags[i];
		if (isSpecIFD (tag.tagId ))
		{
			appendData(buffer, tag.data);
		}
		else if (tag.n > 4)
		{
			// address of new data location to tag
			copyTo (buffer, tag.bufferPtr + 8, topOf(buffer, isLittleEndian));
			// append the data
			if (tag.tagId == TagID_DateTime)
			{
				var d = tag.getString();
				if (d.length > 0)
				{
					// convert date from editor textfield to ISO
					var dateISO = dateStringToISO(d);
					var dateExif = dateStringToExif(dateISO);
					tag.setString(dateExif);
					buffer.add(tag.data);
					// copy date to 0x9003 & 0x9004
					var altDateTag = getTagByIndex(tag.allTags, 0x9003);
					if (altDateTag != null)
					{
 						altDateTag.setString(dateExif);
					}
				}
			}
			else if (tag.tagId == MANUALTAGS)
			{
				var tagStr = tag.getString();
				// split tagStr
				var a = tagStr.split(";");
				for (var j = 0; j < a.length; j++)
				{
					a[j] = a[j].trim();
				}
				tagStr = a[0];
				for (var j = 1; j < a.length; j++)
				{
					tagStr += ";" + a[j];
				}
				tag.setString(tagStr);
				buffer.add(tag.data);
			}
			else
			{
				buffer.add(tag.data);
			}
			if (buffer.length % 2 == 1) buffer.add(0x00);
		}
	}
}

function collectTags(file)
{
	//
	// read exif tags and keep file pointer here
	//
	this.exifBlock = null;
	this.IFDs = null;
	var jpegmarker = marker(file);
	if (jpegmarker.valid && jpegmarker.number == SOI)
	{
		for (;;)
		{
			//Console.writeln("read marker " + toLHex(file.position, 4));
			jpegmarker = marker(file);
			var offset = file.position - 2;
			var dataLen = readLen(file);
			//Console.writeln("marker num  " + toHex(jpegmarker.number));

			if (jpegmarker.number == APP1)
			{
				var baseOffset = file.position + 6;

				file.position = baseOffset;
				// endian flags start here
				var b = file.read( DataType_ByteArray, 2);
				isLittleEndian = b.at(0) == 0x49 && b.at(1) == 0x49;
				// skip 0x002A
				readBytes(file, 2)

				this.IFDs = readIFDs(file, baseOffset, this.IFDs);
				this.exifBlock = {dataLen   : dataLen,
										offset    : offset,
										remainPos : offset + dataLen + 2,
										remainSize: file.size - (offset + dataLen + 2)};
				break;
			}
			else if (jpegmarker.number == APP2)
			{
				//
				// no APP1 in this file, stop @ APP2
				//
				dataLen = 0;
				this.exifBlock = {dataLen   : 0,
										offset    : offset,
										remainPos : offset,
										remainSize: file.size - offset};
				file.position = offset;
				break;
			}
			else if (jpegmarker.number == DQT)
			{
				return;
			}
			else
			{
				file.position += dataLen - 2;
			}
		}
	}
	else
	{
		throw ("not a jpeg file");
	}
}

function readIFDs(file, baseOffset, IFDs)
{
	// file position @ data part

	//
	// read main IFDs of APP1 and return with collection
	// attach special IFDs to tag data
	//
 	var ifdArray = [];
 	var nextIFD = readInt(file, isLittleEndian);

	while (nextIFD != 0)
	{
		file.position = baseOffset + nextIFD;
		var tags = readIFD(file, baseOffset, IFDs);
		ifdArray.push(tags);
		for (var i in tags)
		{
			var tag = tags[i];
			if (isSpecIFD (tag.tagId ))
			{
				// chain specific IDs
				var pos = file.position;
				file.position = tag.tagPtr + 8;
				var ifdPtr = readInt(file, isLittleEndian);
				file.position = ifdPtr + baseOffset;
				var xtags = readIFD(file, baseOffset);
				tag.data = xtags;
				file.position = pos;
			}
		}
		//
		// read offset to next IFD or 0
		//
		nextIFD = readInt(file, isLittleEndian);
	}
	return ifdArray;
}

function readIFD(file, baseOffset, IFDs)
{
	// all byte conversions must follow the endian order
	var tags = [];
	var fieldCount = readShort(file, isLittleEndian);
	for (var i = 0; i < fieldCount; i++)
	{
		var tag = readTag(file, baseOffset, isLittleEndian);
		//Console.writeln("tag "+toLHex(tag.tagId,4)+'\t'+tag.key+'\t'+toLHex(tag.tagPtr,4));
		var all = serializeIFDs(IFDs);
		if (!checkTag(tag, all)) continue;
		if (tag == null) break;
		tags.push(tag);
		//
		// Exif IFD Pointer ?
		//
		if (isSpecIFD (tag.tagId ))
		{
			var nextIFD = baseOffset + tag.dataToInt();
			var pos = file.position;
			// point to next ifd, this ifd has no image data
			file.position = nextIFD;
			tag.data = readIFD(file, baseOffset);
			file.position = pos;
		}
	}
	return tags;
}

function checkTag(tag, all)
{
	// write filter function here
	return true;
}

function readTag(file, baseOffset)//, isLittleEndian)
{
	var pos = file.position;
	var tagId = readShort(file, isLittleEndian);
	//if (tagId == 0) return null;
	var type  = readShort(file, isLittleEndian);
	var count = readInt(file, isLittleEndian);
	var n     = 0;
	/*
 	* 1 = BYTE An 8-bit unsigned integer.,
 	* 2 = ASCII An 8-bit byte containing one 7-bit ASCII code.
	* 		The final byte is terminated with NULL.,
	* 3 = SHORT A 16-bit (2-byte) unsigned integer,
	* 4 = LONG A 32-bit (4-byte) unsigned integer,
	* 5 = RATIONAL Two LONGs.
	* 		The first LONG is the numerator
	* 		and the second LONG expresses the denominator.,
	* 7 = UNDEFINED An 8-bit byte that can take any value depending
	* 		on the field definition,
	* 9 = SLONG A 32-bit (4-byte) signed integer (2's complement notation),
	*10 = SRATIONAL Two SLONGs. The first SLONG is the numerator
	* 		and the second SLONG is the denominator.
	*11 = float 4 bytes
	*12 = float 8 bytes
*/
	if (type == 1)
		n = count;
	else if (type == 2 || type == 7)
		n = count;
	else if (type == 3)
		n = count * 2;
	else if (type == 4 || type == 9 || type == 11)
		n = count * 4;
	else if (type == 5 || type == 10 || type == 12)
		n = count * 8;

	// value or offset
	var data = null;
	var dataEndPtr = 0;

	if (isSpecIFD (tagId )) 	// Exif-specific IFD
	{
  		data = file.read(DataType_ByteArray, n);
		dataEndPtr = file.position;
	}
	else
	{
		if (n == 0)
			data = new ByteArray();
		else
		{
			if (n < 5)
			{
				data = file.read(DataType_ByteArray, 4);
				dataEndPtr = file.position;
			}
			else
			{
				var nextPos = file.position + 4;
 				var p = readInt(file, isLittleEndian);
 				file.position = baseOffset + p;
 				data = file.read(DataType_ByteArray, n);
				dataEndPtr = file.position;
				file.position = nextPos;
			}
		}
	}
	//Console.writeln("read tag "+toLHex(pos, 4) + '\t'
	//+toLHex(tagId, 4) + '\t' + type + '\t' + count + '\t' + toLHex(n, 4) );

	var tag = new exifTag(tagId, type, "TAG" + tagId);
	tag.type = type;
	tag.count = count;
	tag.n = n;
	tag.data = data;
	tag.isLittleEndian = isLittleEndian;
	tag.dataEndPtr = dataEndPtr;
	tag.tagPtr = pos;
	return tag;
}
/* ****************************************************************************
 * Exif TAG object
 * ****************************************************************************
*/
function exifTag(tagId, type, key)//, type, count, n, data, isLittleEndian)
{
	var tag    				= this;
	this.tagId 				= tagId;
	this.key   				= key;
	this.type  				= type;  	// type 1 or 2
	this.count 				= 0;
	this.n     				= 0;
	this.isLittleEndian 	= false;		// isLittleEndian;
	this.edit  				= null;		// edit control
	this.bufferPtr 		= 0;			// position of this tag in buffer
	this.dataEndPtr 		= 0;			// position of this tag's data in file
	this.tagPtr 			= 0;			// position of this tag in file
	this.allTags			= null;		// serialized collection of all tags

	this.data = new ByteArray(4);

	this.setString = function(text)
	{
		if (this.type == 1)
		{
			this.data = stringToShortArray(text);
			this.count = this.data.length;
			this.n     = this.data.length;
		}
		else if (this.type == 2)
		{
			this.data = stringTobyteArray(text);
			this.count = this.data.length;
			this.n     = this.data.length;
		}

		if (this.edit != null)
		{
			this.edit.text =  this.getString();
		}
	}

	this.attach = function(editor)
	{
		this.key   = editor.key;
 		this.edit  = editor.edit;
 		this.edit.text = this.getString();
		this.edit.enabled = true;

		this.edit.onTextUpdated = function( text )
		{
			if (text == "")
			{
				tag.data = new ByteArray();
			}
			else
			{
				if (tag.type == 1)
				{
					tag.data  = stringToShortArray(text);
					tag.count = tag.data.length;
				}
				else if (tag.type == 2)
				{
					tag.data = stringTobyteArray(text);
					tag.count = tag.data.length;
				}
			}
			tag.n = tag.data.length;
		}
	}

	this.getString = function()
	{
		if (this.type == 1)
		{
			return shortArrayToString(this.data);
		}
		else if (this.type == 2)
		{
			return this.data.toString();
		}
		return "";
	}

	this.dataToInt = function()
	{
		var b = this.data.toUint8Array(0, 4);
		var a = 0;
		if (this.isLittleEndian)
		{
			for (var i = 0; i < 4; i++)
			{
				a *= 256;
				a += b[3 - i];
			}
		}
		else
		{
			for (var i = 0; i < 4; i++)
			{
				a *= 256;
				a += b[i];
			}
		}
		return a;
	}

	this.addTag = function(buffer)
	{
		this.bufferPtr = buffer.length;
		var tb = new ByteArray();
		// tagId
		tb.add (numberToByteArray(this.tagId, 2, isLittleEndian));
		// type
		tb.add (numberToByteArray(this.type,  2, isLittleEndian));
		// count
		tb.add (numberToByteArray(this.count, 4, isLittleEndian));
		// prepare data
		tb.add(new ByteArray(4));
		if (!isSpecIFD(this.tagId))
		{
			//
			// move data to buffer, if fits into 4 bytes
			//
			if (this.type == 2)
			{
				if (this.count < 5)
				{
					copyTo (tb, 8, this.data);
				}
			}
			else if (this.n < 5)
			{
				copyTo (tb, 8, this.data);
			}
		}
		buffer.add(tb);
	}
}		// End function exifTag(,,) ***************************************

function topOf(buffer, isLittleEndian)
{
	var c = numberToByteArray(buffer.length, 4, isLittleEndian);
	return c;
}

function shortArrayToString(a)
{
	var b = new ByteArray();
	var i = 0;
	for (;;)
	{
		if (i >= a.length) break;
		b.add(a.at(i));
		i += 2;
	}
	b.add(0x00);
	return b.toString();
}

function stringTobyteArray(a)
{
	var b = new ByteArray();
	for (var i = 0; i < a.length; i++) b.add(a.charAt(i));
	b.add(0x00);
	return b;
}

function stringToShortArray(a)
{
	var b = new ByteArray();
	for (var i = 0; i < a.length; i++)
	{
		b.add(a.charCodeAt(i));
		b.add(0x00);
	}
	b.add(0x00);
	b.add(0x00);
	return b;
}

function copyTo (buffer, index, data)
{
	// copy data to buffer indexed starting at index
	for (var i = 0; i < data.length; i++)
	{
		var j = index + i;
		if (j < buffer.length)
			buffer.at(j, data.at(i));
		else
			break;
	}
}

function byteArrayToInt(byteArray, isLittleEndian)
{
	var b = byteArray.toUint8Array(0, 4);
	var a = 0;
	if (isLittleEndian)
	{
		for (var i = 0; i < 4; i++)
		{
			a *= 256;
			a += b[3 - i];
		}
	}
	else
	{
		for (var i = 0; i < 4; i++)
		{
			a *= 256;
			a += b[i];
		}
	}
	return a;
}

function addBytesToArray(byteArray, bytes)
{
	for (var i = 0; i < bytes.length; i++) byteArray.add(bytes[i]);
}

function numberToByteArray(n, len, isLittleEndian)
{
	var b = new ByteArray(len);
	for (var i = 0; i < len; i++)
	{
		b.at(i, n & (255));
		n = n >> 8;
	}
	if (!isLittleEndian) b.reverse();
	return b;
}

function readInt(file, isLittleEndian)
{
	var a = readBytes(file, 4);
	var k = 0;
	if (isLittleEndian) a = reverse(a);
	for (var i = 0; i < 4; i++)
	{
		k *= 256;
		k += a[i];
	}
	return k;
}

function readShort(file, isLittleEndian)
{
	var a = readBytes(file, 2);
	var k = 0;
	if (isLittleEndian) a = reverse(a);
	for (var i = 0; i < 2; i++)
	{
		var j = a[i];
		k *= 256;
		k += j;
	}
	return k;
}

function readBytes(file, count)
{
	var b = file.read( DataType_ByteArray, count);
   var a = b.toUint8Array()
	return a;
}

function reverse(a)
{
	var m = a.length - 1;
	var j = Math.floor(a.length / 2);
	for (var i = 0; i < j; i++)
	{
		var u = a[m - i];
		a[m - i] = a[i];
		a[i] = u;
	}
	return a;
}

function marker(file)
{
	var b = readBytes(file, 2);
	this.valid = b[0] == 0xFF;
	this.number = b[1];
	return this;
}


function readLen(file)
{
	// big endian
	var b = file.read( DataType_ByteArray, 2);
	return b.at(0) * 256 + b.at(1);
}

function strOf(b, offset, len)
{
  var result = "";
  for (var i = 0; i < len; i++)
  {
  	if (b[i + offset] == 0) break;
    result += String.fromCharCode( b[i + offset] );
  }
  return result;
}

function toHex(d) {
    return  ("0"+(Number(Math.abs(d)).toString(16))).slice(-2).toUpperCase()
}

function serializeIFDs(IFDs)
{
	var tags = [];
	for (var i in IFDs)
	{
		var ifd = tagsCollection(IFDs[i]);
		for (var j in ifd) tags.push(ifd[j]);
	}
	for (var i in tags) tags[i].allTags = tags;
	return tags;
}

function tagsCollection(ifd)
{
	var tags = [];
	for (var i in ifd)
	{
		var tag = ifd[i];
		tags.push(tag);
		if (isSpecIFD (tag.tagId ))	// Exif-specific IFD
		{
			var coll = tagsCollection(tag.data);
			for (var j in coll) tags.push(coll[j]);
		}
	}
	return tags;
}

function toLHex(d, len) {
	// integer to Hex

	var cb = new convert();
	var result = cb.ConvertBase.dec2hex(d);
	while (result.length < len) result = "0" + result;
 	return result.toUpperCase();
}

function dateToExifString(filename)
{
   var fInfo = new FileInfo(filename);
	var d = fInfo.lastModified;
	var s = d.toISOString();
	var a = s.split("T");
	s = a[0] + " ";
	a = a[1].split(".");	// remove .ms
	s += a[0];
	s = s.replace(/-/g, ":");
	return s;
}

function getSetting(key)
{
	var K = key.replace(/ /g,"");
	var temp = Settings.read (SET_PREFIX + "/" + K, DataType_String);
	return temp;
}

function saveSetting(key, value)
{
	var K = key.replace(/ /g,"");
	Settings.write( SET_PREFIX + "/" + K, DataType_String, value );
}

function dump(b, offset)
{
	Console.writeln("dump len  " + toLHex(b.length, 4));
	var l = Math.min(b.length, 4 * 8192);
	var i = 0;
	for (;;)
	{
		Console.write(toLHex(offset + i, 4) + "| ");
		for (var j = 0; j < 16; j++)
		{
			if (j == 8) Console.write("_ ");
			if (i + j < l)
				Console.write(toHex(b[i + j]) + " ");
			else
				Console.write("   ");
		}
		Console.write("| ");
		for (var j = 0; j < 16; j++)
		{
			if (i + j < l)
			{
				var c = b[i + j];
				if (c <= 27)
					Console.write('.');
				else
					Console.write( String.fromCharCode(c) );
			}
			else
				Console.write(" ");
		}
		Console.writeln();
		i += 16;
		if (i > l) break;
	}
}

function dumpByteArray(bArray, offset)
{
	if (bArray == null)
	{
		return;
	}
	else if (bArray.length == 0)
	{
		Console.writeln("Array = " + bArray.length);
		return;
	}
	var d = bArray.toUint8Array();
	dump(d, offset);
}

function dateStringToISO(s)
{
	// input: yyyy-mm-dd [hh[:mm[:ss]]]
	// or     yyyy:mm:dd [hh[:mm[:ss]]]
 	var date = "";
	try
	{
		var a = s.split(" ");
		var b = null;
		a[0] = a[0].replace(/-/g,":");
		b = a[0].split(':');
 		var y = parseInt(b[0]);
		var m = parseInt(b[1]);
		var d = parseInt(b[2]);
		var hh = 0;
		var mm = 0;
		var ss = 0;
		if (a.length > 1)
		{
			b = a[1].split(":");
 			if (b.length > 0) hh = parseInt(b[0]);
			if (b.length > 1) mm = parseInt(b[1]);
			if (b.length > 2) ss = parseInt(b[2]);
		}
		date = pad(y, 4) + '-' + pad(m, 2) + '-' + pad(d, 2) + 'T' +
				 pad(hh, 2) + ':' + pad(mm, 2) + ':' + pad(ss, 2);
	}
	catch (err)
	{
		Console.writeln("dateStringToISO: " + s + ", error: " + err);
	}
	return date;
}

function dateStringToExif(dateISO)
{
	// input: yyyy-mm-ddThh:mm:ss
	// output: yyyy:mm:dd hh:mm:ss
	var date = "";
	try
	{
 		var u  = new Date( dateISO );
		var v = u.valueOf();
		var z = u.getTimezoneOffset();
		u = new Date(v - z * 60 * 1000);
		var s  = u.toISOString();
 		var i  = s.indexOf('.');
		if (i > 1) s = s.substring(0, i);

			var a = s.split("T");
			var b = a[0].split("-");
			var y = parseInt(b[0]);
			var m = parseInt(b[1]);
			var d = parseInt(b[2]);
			var hh = 0;
			var mm = 0;
			var ss = 0;
			if (a.length > 1)
			{
				b = a[1].split(":");
 				if (b.length > 0) hh = parseInt(b[0]);
				if (b.length > 1) mm = parseInt(b[1]);
				if (b.length > 2) ss = parseInt(b[2]);
			}
			date = pad(y, 4) + ':' + pad(m, 2) + ':' + pad(d, 2) + ' ' +
					 pad(hh, 2) + ':' + pad(mm, 2) + ':' + pad(ss, 2);
	}
	catch (err)
	{
		Console.writeln("dateStringToExif: " + s + ", error: " + err);
	}
	return date;
}


function pad(num, size) {
    var s = num + "";
    while (s.length < size) s = "0" + s;
    return s;
}

function convert()
{

    var ConvertBase = function (num) {
        return {
            from : function (baseFrom) {
                return {
                    to : function (baseTo) {
                        return parseInt(num, baseFrom).toString(baseTo);
                    }
                };
            }
        };
    };

    // binary to decimal
    ConvertBase.bin2dec = function (num) {
        return ConvertBase(num).from(2).to(10);
    };

    // binary to hexadecimal
    ConvertBase.bin2hex = function (num) {
        return ConvertBase(num).from(2).to(16);
    };

    // decimal to binary
    ConvertBase.dec2bin = function (num) {
        return ConvertBase(num).from(10).to(2);
    };

    // decimal to hexadecimal
    ConvertBase.dec2hex = function (num) {
        return ConvertBase(num).from(10).to(16);
    };

    // hexadecimal to binary
    ConvertBase.hex2bin = function (num) {
        return ConvertBase(num).from(16).to(2);
    };

    // hexadecimal to decimal
    ConvertBase.hex2dec = function (num) {
        return ConvertBase(num).from(16).to(10);
    };
    this.ConvertBase = ConvertBase;
}

function message(txt)
{
   var bx = new MessageBox(txt, "Error");
   bx.execute();
}

showDialog.prototype = new Dialog;

function main()
{
	Console.hide();
	//Console.clear();
	var dialog = new showDialog();
	dialog.execute();

}

main();
